/*
 * apple2fs driver for apple2 ucsd disk file system access
 * copyright (c)2003 GPL Oskar Schirmer (oskar@scara.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/string.h>
#include "apple2fs.h"

const char *const subdirnames[] = {
	".text",
};

#if APPLE2FS_FILENLEN != 15
#error APPLE2FS_FILENLEN must be 15
#endif
static int apple2fs_subdirname(struct super_block *sb, int n,
	char *sdn)
{
	const char *s = subdirnames[n-1];
	if (APPLE2FS_SBU_INFO(sb, subdirlong[n-1])) {
		memcpy(sdn, &s[1], 4);
		memcpy(&sdn[4], &s[1], 4);
		memcpy(&sdn[8], sdn, 8);
		return APPLE2FS_FILENLEN+1;
	} else {
		memcpy(sdn, s, 5);
		return 5;
	}
}

static inline __u16 fromle16(unsigned char *d)
{
	return (d[1] << 8) | d[0];
}

/*
 * from the root dir get the n-th entry, return it in dire.
 * on success, return 0, otherwise <0
 */
int apple2fs_getrootdirentry(struct super_block *sb, int filenr,
	struct apple2fs_direntry *dire)
{
	struct buffer_head *bh;
	int len = APPLE2FS_SIZEOF_DIRENTRY;
	int ndx = APPLE2FS_ROOTDIR * APPLE2FS_BLOCKSIZE
		+ filenr * APPLE2FS_SIZEOF_DIRENTRY;
	int blk = ndx >> sb->s_blocksize_bits;
	unsigned char rawdire[APPLE2FS_SIZEOF_DIRENTRY];
	ndx &= (sb->s_blocksize - 1);
	while (len > 0) {
		int i = len;
		if (i > sb->s_blocksize - ndx) {
			i = sb->s_blocksize - ndx;
		}
		bh = bread(sb->s_dev, blk, sb->s_blocksize);
		if (!bh) {
			printk("APPLE2FS: read fault on root dir\n");
			return -1;
		}
		memcpy(&rawdire[APPLE2FS_SIZEOF_DIRENTRY-len],
			((char *)bh->b_data) + ndx, i);
		brelse(bh);
		blk += 1;
		ndx = 0;
		len -= i;
	}
	dire->firstblk = fromle16(&rawdire[APPLE2FS_DE_FIRSTBLK]);
	dire->followblk = fromle16(&rawdire[APPLE2FS_DE_FOLLOWBLK]);
	dire->filetype = fromle16(&rawdire[APPLE2FS_DE_FILETYPE]);
	if (dire->filetype == APPLE2FS_FTYPE_DIR) {
		dire->u.dir.nblocks = fromle16(&rawdire[APPLE2FS_DE_NBLOCKS]);
		dire->u.dir.nfiles = fromle16(&rawdire[APPLE2FS_DE_NFILES]);
		dire->u.dir.loadtime = fromle16(&rawdire[APPLE2FS_DE_LOADTIME]);
		dire->u.dir.lastbootdate =
			fromle16(&rawdire[APPLE2FS_DE_LASTBOOTDATE]);
		len = APPLE2FS_DIRNLEN;
	} else {
		dire->u.file.usedbytesinlastblk =
			fromle16(&rawdire[APPLE2FS_DE_USEDBYTESINLASTBLK]);
		dire->u.file.moddate = fromle16(&rawdire[APPLE2FS_DE_MODDATE]);
		len = APPLE2FS_FILENLEN;
	}
	if (len > rawdire[APPLE2FS_DE_NLEN]) {
		len = rawdire[APPLE2FS_DE_NLEN];
	}
	dire->u.file.filenlen = len;
	while (--len >= 0) {
		dire->u.file.filename[len] =
			(rawdire[APPLE2FS_DE_NAME+len] == '/') ? '_' :
				rawdire[APPLE2FS_DE_NAME+len];
	}
	return 0;
}

static int apple2fs_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct apple2fs_direntry dire;
	int offset = 0;
	unsigned int dirno = inode->i_ino - APPLE2FS_ROOT_INO;
	int n;

	if (dirno >= APPLE2FS_INO_FILE(0)) {
		return -ENOTDIR;
	}

	if (filp->f_pos <= 0) {
		if (filldir(dirent, ".", 1, offset++,
				APPLE2FS_ROOT_INO + dirno, DT_DIR)) {
			return 0;
		}
		filp->f_pos += 1;
	}
	if (filp->f_pos <= 1) {
		if (filldir(dirent, "..", 2, offset++,
				APPLE2FS_ROOT_INO, DT_DIR)) {
			return 0;
		}
		filp->f_pos += 1;
	}
	if (dirno == 0) {
		n = filp->f_pos - 1;
		while (n < APPLE2FS_INO_FILE(0)) {
			char sdn[APPLE2FS_FILENLEN+1];
			int len;
			len = apple2fs_subdirname(sb, n, &sdn[0]);
			if (filldir(dirent, sdn, len, offset++,
					APPLE2FS_ROOT_INO + n, DT_DIR)) {
				return 0;
			}
			filp->f_pos += 1;
			n += 1;
		}
	}
	n = filp->f_pos - (dirno == 0 ? APPLE2FS_INO_FILE(0) : 1);
	while (n <= APPLE2FS_SBU_INFO(sb,nfiles)) {
		if (apple2fs_getrootdirentry(sb, n, &dire) < 0) {
			return 0;
		}
		if ((dirno == APPLE2FS_TEXT_INO) ?
		    (dire.filetype == APPLE2FS_FTYPE_TEXT) : 1) {
			if (filldir(dirent, dire.u.file.filename,
					dire.u.file.filenlen, offset,
					APPLE2FS_INO_FILE(n) + dirno, DT_REG)) {
				return 0;
			}
		}
		filp->f_pos += 1;
		n += 1;
	}
	return 0;
}

struct file_operations apple2fs_dir_operations = {
	read:		generic_read_dir,
	readdir:	apple2fs_readdir,
};

static ino_t apple2fs_inode_by_name(struct inode *dir, struct dentry *dentry)
{
	const char *name = dentry->d_name.name;
	int namelen = dentry->d_name.len;
	struct super_block *sb = dir->i_sb;
	int dirno = dir->i_ino - APPLE2FS_ROOT_INO;
	struct apple2fs_direntry dire;
	int n;

	if ((namelen == 1) && (name[0] == '.')) {
		return dir->i_ino;
	}
	if ((namelen == 2) && (name[0] == '.') && (name[1] == '.')) {
		return dentry->d_parent->d_parent->d_inode->i_ino;
	}
	if (dirno == 0) {
		for (n = 1; n < APPLE2FS_INO_FILE(0); n++) {
			char sdn[APPLE2FS_FILENLEN+1];
			int len;
			len = apple2fs_subdirname(sb, n, &sdn[0]);
			if ((len == namelen)
			 && (!memcmp(sdn, name, namelen))) {
				return APPLE2FS_ROOT_INO + n;
			}
		}
	}

	for (n = 1; n <= APPLE2FS_SBU_INFO(sb,nfiles); n++) {
		if (apple2fs_getrootdirentry(sb, n, &dire) < 0) {
			return 0;
		}
		if ((dirno == APPLE2FS_TEXT_INO) ?
		    (dire.filetype == APPLE2FS_FTYPE_TEXT) : 1) {
			if ((dire.u.file.filenlen == namelen)
			 && (!memcmp(dire.u.file.filename, name, namelen))) {
				return APPLE2FS_INO_FILE(n) + dirno;
			}
		}
	}
	return 0;
}

static struct dentry *apple2fs_lookup(struct inode *dir, struct dentry *dentry)
{
	struct inode *inode;
	ino_t ino;

	if (dentry->d_name.len > APPLE2FS_FILENLEN+1)
		return ERR_PTR(-ENAMETOOLONG);

	ino = apple2fs_inode_by_name(dir, dentry);
	inode = NULL;
	if (ino) {
		inode = iget(dir->i_sb, ino);
		if (!inode) {
			return ERR_PTR(-EACCES);
		}
	}
	d_add(dentry, inode);
	return NULL;
}

struct inode_operations apple2fs_dir_inode_operations =
{
	lookup:		apple2fs_lookup,
};
