/*
 * apple2fs driver for apple2 ucsd disk file system access
 * copyright (c)2003 GPL Oskar Schirmer (oskar@scara.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <asm/uaccess.h>
#include "apple2fs.h"

static int apple2fs_open_file(struct inode *inode, struct file *filp)
{
	struct apple2fs_file_info *pd;
	pd = (struct apple2fs_file_info *)kmalloc(
			sizeof(struct apple2fs_file_info), GFP_KERNEL);
	if (pd == NULL) {
		return -ENOMEM;
	}
	if ((inode->i_ino >= APPLE2FS_INO_FILE(1))
	 && ((inode->i_ino & APPLE2FS_INO_MASK) == APPLE2FS_TEXT_INO)) {
		pd->phys_pos = APPLE2FS_TEXTPHYSBEGIN;
	} else {
		pd->phys_pos = 0;
	}
	pd->last_fpos = 0;
	pd->leftover_dle = 0;
	pd->accum_null = 0;
	filp->private_data = pd;
	return 0;
}

static int apple2fs_release_file(struct inode *inode, struct file *filp)
{
	kfree (filp->private_data);
	return 0;
}

int apple2fs_scantextfile(struct super_block *sb,
	struct apple2fs_file_info *pd,
	int firstblock, int count, char *buf)
{
	int left;
	int ndx = (firstblock << APPLE2FS_BLOCKSIZE_BITS) + pd->phys_pos;
	int blk = ndx >> sb->s_blocksize_bits;
	struct buffer_head *bh;
	int i = 0;
	ndx &= sb->s_blocksize - 1;
	left = sb->s_blocksize - ndx;
	while (count < 0 ? pd->phys_pos < -count : i < count) {
		char *c;
		bh = bread(sb->s_dev, blk, sb->s_blocksize);
		if (!bh) {
			printk("APPLE2FS: read fault on file\n");
			return -EFAULT;
		}
		c = ((char *)bh->b_data) + ndx;
		do {
			while ((pd->leftover_dle > 0)
			    && ((count < 0) || (i < count))) {
				pd->leftover_dle -= 1;
				if (buf != NULL) {
					put_user (' ', buf);
					buf += 1;
				}
				i += 1;
			}
			if ((count < 0) || (i < count)) {
				char d = *c;
				if (pd->leftover_dle < 0) {
					pd->leftover_dle = (d - ' ') & 0xFF;
					c += 1;
					left -= 1;
				} else if (d == 0) {
					c += 1;
					left -= 1;
					if ((pd->phys_pos + sb->s_blocksize
							- ndx - left)
						% APPLE2FS_BLOCKSIZE == 0) {
						pd->accum_null = 0;
					} else {
						pd->accum_null += 1;
					}
				} else {
					while ((pd->accum_null > 0)
					   && ((count < 0) || (i < count))) {
						if (buf != NULL) {
							put_user(0, buf);
							buf += 1;
						}
						pd->accum_null -= 1;
						i += 1;
					}
					if ((count < 0) || (i < count)) {
						switch (d) {
						case 16:
							pd->leftover_dle = -1;
							break;
						case 13:
							d = 10;
						default:
							if (buf != NULL) {
								put_user(d,buf);
								buf += 1;
							}
							i += 1;
							break;
						}
						c += 1;
						left -= 1;
					}
				}
			}
		} while (((left > 0) || (pd->leftover_dle > 0)
				     || (pd->accum_null > 0))
		      && ((count < 0) || (i < count)));
		brelse(bh);
		pd->phys_pos += sb->s_blocksize - ndx - left;
		blk += 1;
		ndx = 0;
		left = sb->s_blocksize;
	}
	pd->last_fpos += i;
	return i;
}

static ssize_t apple2fs_read_file(struct file *filp,
	char *buf,
	size_t count,
	loff_t *ppos)
{
	struct apple2fs_file_info *pd =
		(struct apple2fs_file_info *)filp->private_data;
	struct inode *inode = filp->f_dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct buffer_head *bh;
	if (count > inode->i_size - filp->f_pos) {
		count = inode->i_size - filp->f_pos;
	}
	if (count <= 0) {
		return 0;
	}
	if ((inode->i_ino & APPLE2FS_INO_MASK) == APPLE2FS_TEXT_INO) {
		int r;
		if (filp->f_pos < pd->last_fpos) {
			pd->phys_pos = APPLE2FS_TEXTPHYSBEGIN;
			pd->last_fpos = 0;
			pd->leftover_dle = 0;
			pd->accum_null = 0;
		}
		if (filp->f_pos > pd->last_fpos) {
			r = apple2fs_scantextfile (sb, pd,
				APPLE2FS_IPU_INFO(inode,firstblock),
				filp->f_pos - pd->last_fpos, NULL);
			filp->f_pos = pd->last_fpos;
			if (r < 0) {
				return r;
			}
		}
		r = apple2fs_scantextfile (sb, pd, 
			APPLE2FS_IPU_INFO(inode,firstblock), count, buf);
		filp->f_pos = pd->last_fpos;
		return r;
	} else {
		int left;
		int ndx = (APPLE2FS_IPU_INFO(inode,firstblock)
			<< APPLE2FS_BLOCKSIZE_BITS) + pd->phys_pos;
		int blk = ndx >> sb->s_blocksize_bits;
		ndx &= sb->s_blocksize - 1;
		left = sb->s_blocksize - ndx;
		bh = bread(sb->s_dev, blk, sb->s_blocksize);
		if (!bh) {
			printk("APPLE2FS: read fault on file\n");
			return -EFAULT;
		}
		if (count > left) {
			count = left;
		}
		if (copy_to_user (buf, ((char *)bh->b_data) + ndx, count)) {
			printk ("APPLE2FS: copy_to_user failed\n");
			brelse(bh);
			return -EFAULT;
		}
		brelse(bh);
		pd->phys_pos += count;
		filp->f_pos = pd->phys_pos;
		return count;
	}
}

struct file_operations apple2fs_file_operations = {
	llseek:		generic_file_llseek,
	read:		apple2fs_read_file,
	mmap:		generic_file_mmap,
	open:		apple2fs_open_file,
	release:	apple2fs_release_file,
};
