/*
 * apple2fs driver for apple2 ucsd disk file system access
 * copyright (c)2003 GPL Oskar Schirmer (oskar@scara.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/config.h>
#include <linux/module.h>

#include <linux/blkdev.h>
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include "apple2fs.h"

static int apple2fs_statfs(struct super_block *sb, struct statfs *buf)
{
	buf->f_type = 0x41504C32;
	buf->f_bsize = sb->s_blocksize;
	buf->f_blocks = APPLE2FS_SBU_INFO(sb,nblocks);
	buf->f_bfree = 0; /* not true, but RO */
	buf->f_bavail = 0;
	buf->f_files = APPLE2FS_SBU_INFO(sb,nfiles);
	buf->f_ffree = 0;
	buf->f_namelen = APPLE2FS_FILENLEN+1;
	return 0;
}

static void apple2fs_put_super(struct super_block *sb)
{
	kfree(sb->u.generic_sbp);
}

static struct super_operations apple2fs_sops = {
	read_inode:	apple2fs_read_inode,
	put_inode:	apple2fs_put_inode,
	put_super:	apple2fs_put_super,
	statfs:		apple2fs_statfs
};

static struct super_block *apple2fs_read_super(struct super_block *sb,
	void *data,
	int silent)
{
	kdev_t dev = sb->s_dev;
	struct apple2fs_direntry dire;
	int blocksize;
	int n, b;

	blocksize = get_hardsect_size(dev);
	if (blocksize < APPLE2FS_BLOCKSIZE) {
		blocksize = APPLE2FS_BLOCKSIZE;
	}

	if (set_blocksize(dev, blocksize) < 0) {
		printk("APPLE2FS: unable to set blocksize %d\n", blocksize);
		return NULL;
	}

	sb->u.generic_sbp = (void *)kmalloc(
			sizeof(struct apple2fs_sb_info), GFP_KERNEL);
	if (sb->u.generic_sbp == NULL) {
		printk("APPLE2FS: kmalloc failed for superblock\n");
		return NULL;
	}

	sb->s_blocksize_bits = APPLE2FS_BLOCKSIZE_BITS;
	sb->s_blocksize = blocksize;
	while (blocksize > APPLE2FS_BLOCKSIZE) {
		sb->s_blocksize_bits += 1;
		blocksize >>= 1;
	}
	if (apple2fs_getrootdirentry(sb, 0, &dire) < 0) {
		kfree(sb->u.generic_sbp);
		return NULL;
	}

	APPLE2FS_SBU_INFO(sb,nblocks) = dire.u.dir.nblocks;
	sb->s_maxbytes = APPLE2FS_BLOCKSIZE * APPLE2FS_SBU_INFO(sb,nblocks);
	APPLE2FS_SBU_INFO(sb,nfiles) = dire.u.dir.nfiles;
	APPLE2FS_SBU_INFO(sb,firstdatablk) = b = dire.followblk;
	memset(&APPLE2FS_SBU_INFO(sb,subdirlong[0]), 0, APPLE2FS_INO_MASK);

	/* check root dir for validity */
	if ((APPLE2FS_SIZEOF_DIRENTRY*(APPLE2FS_SBU_INFO(sb,nfiles)+1) >
		APPLE2FS_BLOCKSIZE
			*(APPLE2FS_SBU_INFO(sb,firstdatablk)-APPLE2FS_ROOTDIR))
	 || (dire.firstblk != 0)
	 || (APPLE2FS_SBU_INFO(sb,firstdatablk) <= APPLE2FS_ROOTDIR)
	 || (APPLE2FS_SBU_INFO(sb,firstdatablk) >=
		APPLE2FS_SBU_INFO(sb,nblocks))
	 || (dire.filetype != APPLE2FS_FTYPE_DIR)
	 || (dire.u.dir.dirnlen > APPLE2FS_DIRNLEN)) {
		printk("APPLE2FS: invalid root dir [%d]\n", 0);
		kfree(sb->u.generic_sbp);
		return NULL;
	}
	for (n = 1; n <= APPLE2FS_SBU_INFO(sb,nfiles); n++) {
		int i;
		if (apple2fs_getrootdirentry(sb, n, &dire) < 0) {
			kfree(sb->u.generic_sbp);
			return NULL;
		}
		if ((dire.firstblk < b)
		 || (dire.firstblk > dire.followblk)
		 || (dire.followblk > APPLE2FS_SBU_INFO(sb,nblocks))
		 || (dire.filetype >= APPLE2FS_NR_FTYPE)
		 || (dire.filetype == APPLE2FS_FTYPE_DIR)
		 || (dire.u.file.filenlen > APPLE2FS_FILENLEN)
		 || (dire.u.file.usedbytesinlastblk > APPLE2FS_BLOCKSIZE)) {
			printk("APPLE2FS: invalid root dir [%d]\n", n);
			kfree(sb->u.generic_sbp);
			return NULL;
		}
		b = dire.followblk;
		for (i = 0; i < APPLE2FS_INO_MASK; i++) {
			if ((dire.u.file.filenlen == 5)
			 && (!memcmp(dire.u.file.filename,subdirnames[i],5))) {
				APPLE2FS_SBU_INFO(sb,subdirlong[i]) = 1;
			}
		}
	}

	sb->s_flags |= MS_RDONLY;
	sb->s_op = &apple2fs_sops;
	sb->s_root = d_alloc_root(iget(sb, APPLE2FS_ROOT_INO));
	return sb;
}

static DECLARE_FSTYPE_DEV(apple2fs_type, "apple2fs", apple2fs_read_super);

static int __init init_apple2fs(void)
{
	return register_filesystem(&apple2fs_type);
}

static void __exit exit_apple2fs(void)
{
	unregister_filesystem(&apple2fs_type);
}

module_init(init_apple2fs)
module_exit(exit_apple2fs)
